@@ -5,17 +5,13 @@ except BaseException:
 
 
 class AdjacencyTest(g.unittest.TestCase):
-
     def test_radius(self):
-
         for radius in [0.1, 1.0, 3.1459, 29.20]:
-            m = g.trimesh.creation.cylinder(
-                radius=radius, height=radius * 10)
+            m = g.trimesh.creation.cylinder(radius=radius, height=radius * 10)
 
             # remove the cylinder cap
             signs = (g.np.sign(m.vertices[:, 2]) < 0)[m.faces]
-            not_cap = ~g.np.logical_or(
-                signs.all(axis=1), ~signs.any(axis=1))
+            not_cap = ~g.np.logical_or(signs.all(axis=1), ~signs.any(axis=1))
             m.update_faces(not_cap)
 
             # compare the calculated radius
@@ -25,6 +21,6 @@ class AdjacencyTest(g.unittest.TestCase):
             assert g.np.allclose(radii, radius, atol=radius / 100)
 
 
-if __name__ == '__main__':
+if __name__ == "__main__":
     g.trimesh.util.attach_to_log()
     g.unittest.main()
