@@ -1,28 +1,64 @@
 import io
-import sys
 import uuid
 import zipfile
+from collections import defaultdict
 
-import collections
 import numpy as np
 
-from .. import util
-from .. import graph
-
+from .. import graph, util
 from ..constants import log
-
-try:
-    import networkx as nx
-except BaseException as E:
-    # create a dummy module which will raise the ImportError
-    # or other exception only when someone tries to use networkx
-    from ..exceptions import ExceptionModule
-    nx = ExceptionModule(E)
+from ..util import unique_name
 
 
-def load_3MF(file_obj,
-             postprocess=True,
-             **kwargs):
+def _read_mesh(mesh):
+    """
+    Read a `<mesh ` XML element into Numpy vertices and faces.
+
+    This is generally the most expensive operation in the load as it
+    has to operate in Python-space on every single vertex and face.
+
+    Parameters
+    ----------
+    mesh : lxml.etree.Element
+      Input mesh element with `vertex` and `triangle` children.
+
+    Returns
+    ----------
+    vertex_array : (n, 3) float64
+      Vertices
+    face_array : (n, 3) int64
+      Indexes of vertices forming triangles.
+    """
+    # get the XML elements for vertices and faces
+    vertices = mesh.find("{*}vertices")
+    faces = mesh.find("{*}triangles")
+
+    # get every value as a flat space-delimited string
+    # this is very sensitive as it is large, i.e. it is
+    # much faster with the full list comprehension before
+    # the `.join` as the giant string can be fully allocated
+    vs = " ".join(
+        [
+            f"{i.attrib['x']} {i.attrib['y']} {i.attrib['z']}"
+            for i in vertices.iter("{*}vertex")
+        ]
+    )
+    # convert every value to floating point in one-shot rather than in a loop
+    v_array = np.fromstring(vs, dtype=np.float64, sep=" ").reshape((-1, 3))
+
+    # do the same behavior for faces but as an integer
+    fs = " ".join(
+        [
+            f"{i.attrib['v1']} {i.attrib['v2']} {i.attrib['v3']}"
+            for i in faces.iter("{*}triangle")
+        ]
+    )
+    f_array = np.fromstring(fs, dtype=np.int64, sep=" ").reshape((-1, 3))
+
+    return v_array, f_array
+
+
+def load_3MF(file_obj, postprocess=True, **kwargs):
     """
     Load a 3MF formatted file into a Trimesh scene.
 
@@ -36,104 +72,114 @@ def load_3MF(file_obj,
     kwargs : dict
       Constructor arguments for `trimesh.Scene`
     """
+
     # dict, {name in archive: BytesIo}
-    archive = util.decompress(file_obj, file_type='zip')
+    archive = util.decompress(file_obj, file_type="zip")
     # get model with case-insensitive keys
-    model = next(iter(v for k, v in archive.items()
-                      if '3d/3dmodel.model' in k.lower()))
+    model = next(iter(v for k, v in archive.items() if "3d/3dmodel.model" in k.lower()))
 
     # read root attributes only from XML first
-    event, root = next(etree.iterparse(
-        model, tag=('{*}model'), events=('start',)))
+    _event, root = next(etree.iterparse(model, tag=("{*}model"), events=("start",)))
     # collect unit information from the tree
-    if 'unit' in root.attrib:
-        metadata = {'units': root.attrib['unit']}
+    if "unit" in root.attrib:
+        metadata = {"units": root.attrib["unit"]}
     else:
         # the default units, defined by the specification
-        metadata = {'units': 'millimeters'}
+        metadata = {"units": "millimeters"}
 
     # { mesh id : mesh name}
     id_name = {}
     # { mesh id: (n,3) float vertices}
-    v_seq = {}
+    v_seq = defaultdict(list)
     # { mesh id: (n,3) int faces}
-    f_seq = {}
+    f_seq = defaultdict(list)
     # components are objects that contain other objects
     # {id : [other ids]}
-    components = collections.defaultdict(list)
+    components = defaultdict(list)
     # load information about the scene graph
     # each instance is a single geometry
     build_items = []
 
+    # keep track of names we can use
+    consumed_counts = {}
     consumed_names = set()
+
     # iterate the XML object and build elements with an LXML iterator
     # loaded elements are cleared to avoid ballooning memory
     model.seek(0)
-    for event, obj in etree.iterparse(model, tag=('{*}object', '{*}build')):
+    for _, obj in etree.iterparse(model, tag=("{*}object", "{*}build"), events=("end",)):
         # parse objects
-        if 'object' in obj.tag:
+        if "object" in obj.tag:
             # id is mandatory
-            index = obj.attrib['id']
+            index = obj.attrib["id"]
 
             # start with stored name
-            name = obj.attrib.get('name', str(index))
             # apparently some exporters name multiple meshes
             # the same thing so check to see if it's been used
-            if name in consumed_names:
-                name = name + str(index)
+            name = unique_name(
+                obj.attrib.get("name", str(index)), consumed_names, consumed_counts
+            )
             consumed_names.add(name)
             # store name reference on the index
             id_name[index] = name
 
             # if the object has actual geometry data parse here
-            for mesh in obj.iter('{*}mesh'):
-                vertices = mesh.find('{*}vertices')
-                v_seq[index] = np.array([[i.attrib['x'],
-                                          i.attrib['y'],
-                                          i.attrib['z']] for
-                                         i in vertices.iter('{*}vertex')],
-                                        dtype=np.float64)
-                vertices.clear()
-                vertices.getparent().remove(vertices)
-
-                faces = mesh.find('{*}triangles')
-                f_seq[index] = np.array([[i.attrib['v1'],
-                                          i.attrib['v2'],
-                                          i.attrib['v3']] for
-                                         i in faces.iter('{*}triangle')],
-                                        dtype=np.int64)
-                faces.clear()
-                faces.getparent().remove(faces)
+            for mesh in obj.iter("{*}mesh"):
+                v, f = _read_mesh(mesh)
+                v_seq[index].append(v)
+                f_seq[index].append(f)
 
             # components are references to other geometries
-            for c in obj.iter('{*}component'):
-                mesh_index = c.attrib['objectid']
+            for c in obj.iter("{*}component"):
+                mesh_index = c.attrib["objectid"]
                 transform = _attrib_to_transform(c.attrib)
                 components[index].append((mesh_index, transform))
 
+                # if this references another file as the `path` attrib
+                path = next(
+                    (v.strip("/") for k, v in c.attrib.items() if k.endswith("path")),
+                    None,
+                )
+                if path is not None and path in archive:
+                    archive[path].seek(0)
+                    name = unique_name(
+                        obj.attrib.get("name", str(mesh_index)),
+                        consumed_names,
+                        consumed_counts,
+                    )
+                    consumed_names.add(name)
+                    # store name reference on the index
+                    id_name[mesh_index] = name
+
+                    for _, m in etree.iterparse(
+                        archive[path], tag=("{*}mesh"), events=("end",)
+                    ):
+                        v, f = _read_mesh(m)
+                        v_seq[mesh_index].append(v)
+                        f_seq[mesh_index].append(f)
+
         # parse build
-        if 'build' in obj.tag:
+        if "build" in obj.tag:
             # scene graph information stored here, aka "build" the scene
-            for item in obj.iter('{*}item'):
+            for item in obj.iter("{*}item"):
                 # get a transform from the item's attributes
                 transform = _attrib_to_transform(item.attrib)
                 # the index of the geometry this item instantiates
-                build_items.append((item.attrib['objectid'], transform))
-
-        # free resources
-        obj.clear()
-        obj.getparent().remove(obj)
-        del obj
+                build_items.append((item.attrib["objectid"], transform))
 
     # have one mesh per 3MF object
     # one mesh per geometry ID, store as kwargs for the object
     meshes = {}
     for gid in v_seq.keys():
+        v, f = util.append_faces(v_seq[gid], f_seq[gid])
         name = id_name[gid]
-        meshes[name] = {'vertices': v_seq[gid],
-                        'faces': f_seq[gid],
-                        'metadata': metadata.copy()}
-        meshes[name].update(kwargs)
+        meshes[name] = {
+            "vertices": v,
+            "faces": f,
+            "metadata": metadata.copy(),
+        }
+        # apply any keyword arguments that aren't None
+        meshes[name].update({k: v for k, v in kwargs.items() if v is not None})
 
     # turn the item / component representation into
     # a MultiDiGraph to compound our pain
@@ -141,24 +187,21 @@ def load_3MF(file_obj,
     # build items are the only things that exist according to 3MF
     # so we accomplish that by linking them to the base frame
     for gid, tf in build_items:
-        g.add_edge('world', gid, matrix=tf)
+        g.add_edge("world", gid, matrix=tf)
     # components are instances which need to be linked to base
     # frame by a build_item
     for start, group in components.items():
-        for i, (gid, tf) in enumerate(group):
+        for gid, tf in group:
             g.add_edge(start, gid, matrix=tf)
 
     # turn the graph into kwargs for a scene graph
     # flatten the scene structure and simplify to
     # a single unique node per instance
     graph_args = []
-    parents = collections.defaultdict(set)
-    for path in graph.multigraph_paths(G=g,
-                                       source='world'):
+    parents = defaultdict(set)
+    for path in graph.multigraph_paths(G=g, source="world"):
         # collect all the transform on the path
-        transforms = graph.multigraph_collect(G=g,
-                                              traversal=path,
-                                              attrib='matrix')
+        transforms = graph.multigraph_collect(G=g, traversal=path, attrib="matrix")
         # combine them into a single transform
         if len(transforms) == 1:
             transform = transforms[0]
@@ -169,8 +212,9 @@ def load_3MF(file_obj,
         last = path[-1][0]
         # if someone included an undefined component, skip it
         if last not in id_name:
-            log.debug('id {} included but not defined!'.format(last))
+            log.warning(f"id {last} included but not defined!")
             continue
+
         # frame names unique
         name = id_name[last] + util.unique_id()
         # index in meshes
@@ -181,14 +225,18 @@ def load_3MF(file_obj,
             parent = path[-2][0]
             parents[parent].add(last)
 
-        graph_args.append({'frame_from': 'world',
-                           'frame_to': name,
-                           'matrix': transform,
-                           'geometry': geom})
+        graph_args.append(
+            {
+                "frame_from": "world",
+                "frame_to": name,
+                "matrix": transform,
+                "geometry": geom,
+            }
+        )
 
     # solidworks will export each body as its own mesh with the part
     # name as the parent so optionally rename and combine these bodies
-    if postprocess and all('body' in i.lower() for i in meshes.keys()):
+    if postprocess and all("body" in i.lower() for i in meshes.keys()):
         # don't rename by default
         rename = {k: k for k in meshes.keys()}
         for parent, mesh_name in parents.items():
@@ -197,29 +245,27 @@ def load_3MF(file_obj,
             if len(mesh_name) != 1:
                 continue
             # rename the part
-            rename[id_name[next(iter(mesh_name))]] = id_name[parent].split(
-                '(')[0]
+            rename[id_name[next(iter(mesh_name))]] = id_name[parent].split("(")[0]
 
         # apply the rename operation meshes
         meshes = {rename[k]: m for k, m in meshes.items()}
         # rename geometry references in the scene graph
         for arg in graph_args:
-            if 'geometry' in arg:
-                arg['geometry'] = rename[arg['geometry']]
+            if "geometry" in arg:
+                arg["geometry"] = rename[arg["geometry"]]
 
     # construct the kwargs to load the scene
-    kwargs = {'base_frame': 'world',
-              'graph': graph_args,
-              'geometry': meshes,
-              'metadata': metadata}
+    kwargs = {
+        "base_frame": "world",
+        "graph": graph_args,
+        "geometry": meshes,
+        "metadata": metadata,
+    }
 
     return kwargs
 
 
-def export_3MF(mesh,
-               batch_size=4096,
-               compression=zipfile.ZIP_DEFLATED,
-               compresslevel=5):
+def export_3MF(mesh, batch_size=4096, compression=zipfile.ZIP_DEFLATED, compresslevel=5):
     """
     Converts a Trimesh object into a 3MF file.
 
@@ -240,11 +286,6 @@ def export_3MF(mesh,
       Represents geometry as a 3MF file.
     """
 
-    if sys.version_info < (3, 6):
-        # Python only added 'w' mode to `zipfile` in Python 3.6
-        # and it is not worth the effort to work around
-        raise NotImplementedError(
-            "3MF export requires Python >= 3.6")
     from ..scene.scene import Scene
 
     if not isinstance(mesh, Scene):
@@ -264,12 +305,12 @@ def export_3MF(mesh,
         "sc": "http://schemas.microsoft.com/3dmanufacturing/securecontent/2019/04",
     }
 
-    rels_nsmap = {
-        None: "http://schemas.openxmlformats.org/package/2006/relationships"
-    }
+    rels_nsmap = {None: "http://schemas.openxmlformats.org/package/2006/relationships"}
 
     # model ids
-    def model_id(x, models=[]):
+    models = []
+
+    def model_id(x):
         if x not in models:
             models.append(x)
         return str(models.index(x) + 1)
@@ -278,15 +319,14 @@ def export_3MF(mesh,
     file_obj = io.BytesIO()
 
     # specify the parameters for the zip container
-    zip_kwargs = {'compression': compression}
+    zip_kwargs = {"compression": compression}
     # compresslevel was added in Python 3.7
-    if sys.version_info >= (3, 7):
-        zip_kwargs['compresslevel'] = compresslevel
+    zip_kwargs["compresslevel"] = compresslevel
 
-    with zipfile.ZipFile(file_obj, mode='w', **zip_kwargs) as z:
+    with zipfile.ZipFile(file_obj, mode="w", **zip_kwargs) as z:
         # 3dmodel.model
-        with z.open("3D/3dmodel.model", mode='w') as f, etree.xmlfile(
-                f, encoding="utf-8"
+        with z.open("3D/3dmodel.model", mode="w") as f, etree.xmlfile(
+            f, encoding="utf-8"
         ) as xf:
             xf.write_declaration()
 
@@ -301,20 +341,18 @@ def export_3MF(mesh,
                             "id": model_id(name),
                             "name": name,
                             "type": "model",
-                            "p:UUID": str(uuid.uuid4())
+                            "p:UUID": str(uuid.uuid4()),
                         }
                         with xf.element("object", **attribs):
                             with xf.element("mesh"):
                                 with xf.element("vertices"):
-                                    # vertex nodes are writed directly to the file
+                                    # vertex nodes are written directly to the file
                                     # so make sure lxml's buffer is flushed
                                     xf.flush()
-                                    for i in range(
-                                            0, len(m.vertices), batch_size):
-                                        batch = m.vertices[i: i + batch_size]
+                                    for i in range(0, len(m.vertices), batch_size):
+                                        batch = m.vertices[i : i + batch_size]
                                         fragment = (
-                                            '<vertex x="{}" y="{}" z="{}" />'
-                                            * len(batch)
+                                            '<vertex x="{}" y="{}" z="{}" />' * len(batch)
                                         )
                                         f.write(
                                             fragment.format(*batch.flatten()).encode(
@@ -323,9 +361,8 @@ def export_3MF(mesh,
                                         )
                                 with xf.element("triangles"):
                                     xf.flush()
-                                    for i in range(
-                                            0, len(m.faces), batch_size):
-                                        batch = m.faces[i: i + batch_size]
+                                    for i in range(0, len(m.faces), batch_size):
+                                        batch = m.faces[i : i + batch_size]
                                         fragment = (
                                             '<triangle v1="{}" v2="{}" v3="{}" />'
                                             * len(batch)
@@ -347,7 +384,7 @@ def export_3MF(mesh,
                             "id": model_id(node),
                             "name": node,
                             "type": "model",
-                            "p:UUID": str(uuid.uuid4())
+                            "p:UUID": str(uuid.uuid4()),
                         }
                         with xf.element("object", **attribs):
                             with xf.element("components"):
@@ -378,16 +415,16 @@ def export_3MF(mesh,
                         transform = " ".join(
                             str(i) for i in np.array(data["matrix"])[:3, :4].T.flatten()
                         )
-                        uuid_tag = "{{{}}}UUID".format(model_nsmap['p'])
+                        uuid_tag = "{{{}}}UUID".format(model_nsmap["p"])
                         xf.write(
                             etree.Element(
                                 "item",
                                 {
                                     "objectid": model_id(node),
                                     "transform": transform,
-                                    uuid_tag: str(uuid.uuid4())
+                                    uuid_tag: str(uuid.uuid4()),
                                 },
-                                nsmap=model_nsmap
+                                nsmap=model_nsmap,
                             )
                         )
 
@@ -412,9 +449,7 @@ def export_3MF(mesh,
         ) as xf:
             xf.write_declaration()
             # xml namespaces
-            nsmap = {
-                None: "http://schemas.openxmlformats.org/package/2006/content-types"
-            }
+            nsmap = {None: "http://schemas.openxmlformats.org/package/2006/content-types"}
 
             # stream elements
             types = [
@@ -430,11 +465,7 @@ def export_3MF(mesh,
             ]
             with xf.element("Types", nsmap=nsmap):
                 for ext, ctype in types:
-                    xf.write(
-                        etree.Element(
-                            "Default",
-                            Extension=ext,
-                            ContentType=ctype))
+                    xf.write(etree.Element("Default", Extension=ext, ContentType=ctype))
 
     return file_obj.getvalue()
 
@@ -453,18 +484,22 @@ def _attrib_to_transform(attrib):
     """
 
     transform = np.eye(4, dtype=np.float64)
-    if 'transform' in attrib:
+    if "transform" in attrib:
         # wangle their transform format
-        values = np.array(
-            attrib['transform'].split(),
-            dtype=np.float64).reshape((4, 3)).T
+        values = np.array(attrib["transform"].split(), dtype=np.float64).reshape((4, 3)).T
         transform[:3, :4] = values
     return transform
 
 
 # do import here to keep lxml a soft dependency
 try:
+    import networkx as nx
     from lxml import etree
-    _three_loaders = {'3mf': load_3MF}
-except ImportError:
-    _three_loaders = {}
+
+    _three_loaders = {"3mf": load_3MF}
+    _3mf_exporters = {"3mf": export_3MF}
+except BaseException as E:
+    from ..exceptions import ExceptionWrapper
+
+    _three_loaders = {"3mf": ExceptionWrapper(E)}
+    _3mf_exporters = {"3mf": ExceptionWrapper(E)}
