@@ -1,20 +1,19 @@
-import numpy as np
 import collections
-
 from copy import deepcopy
 
-from .. import util
-from .. import caching
-from .. import transformations
+import numpy as np
 
+from .. import caching, util
 from ..caching import hash_fast
+from ..transformations import fix_rigid, quaternion_matrix, rotation_matrix
+from ..typed import ArrayLike, Hashable, NDArray, Optional, Sequence, Tuple, Union
 
 # we compare to identity a lot
 _identity = np.eye(4)
-_identity.flags['WRITEABLE'] = False
+_identity.flags["WRITEABLE"] = False
 
 
-class SceneGraph(object):
+class SceneGraph:
     """
     Hold data about positions and instances of geometry
     in a scene. This includes a forest (i.e. multi-root tree)
@@ -23,7 +22,7 @@ class SceneGraph(object):
     nodes.
     """
 
-    def __init__(self, base_frame='world'):
+    def __init__(self, base_frame="world", repair_rigid=1e-5):
         """
         Create a scene graph, holding homogeneous transformation
         matrices and instance information about geometry.
@@ -32,11 +31,20 @@ class SceneGraph(object):
         -----------
         base_frame : any
           The root node transforms will be positioned from.
+        repair_rigid : None or float
+          If a float will attempt to repair rotation matrices
+          where `M @ M.T` differs from an identity matrix by
+          more than floating point zero but less than this value.
+          This can happen in a deep tree with a lot of matrix
+          multiplies.
         """
         # a graph structure, subclass of networkx DiGraph
         self.transforms = EnforcedForest()
         # hashable, the base or root frame
         self.base_frame = base_frame
+        # if passed as a float try to repair rigid transforms
+        # that have accumulated floating point error
+        self.repair_rigid = repair_rigid
         # cache transformation matrices keyed with tuples
         self._cache = caching.Cache(self.__hash__)
 
@@ -63,7 +71,7 @@ class SceneGraph(object):
           Distance to translate
         geometry : hashable
           Geometry object name, e.g. 'mesh_0'
-        extras: dictionary
+        metadata: dictionary
           Optional metadata attached to the new frame
           (exports to glTF node 'extras').
         """
@@ -72,21 +80,21 @@ class SceneGraph(object):
             frame_from = self.base_frame
 
         # pass through
-        attr = {k: v for k, v in kwargs.items()
-                if k in {'geometry', 'extras'}}
+        attr = {k: v for k, v in kwargs.items() if k in {"geometry", "metadata"}}
         # convert various kwargs to a single matrix
-        attr['matrix'] = kwargs_to_matrix(**kwargs)
+        attr["matrix"] = kwargs_to_matrix(**kwargs)
 
         # add the edges for the transforms
         # wi ll return if it changed anything
         self.transforms.add_edge(frame_from, frame_to, **attr)
 
         # set the node attribute with the geometry information
-        if 'geometry' in kwargs:
-            self.transforms.node_data[
-                frame_to]['geometry'] = kwargs['geometry']
+        if "geometry" in kwargs:
+            self.transforms.node_data[frame_to]["geometry"] = kwargs["geometry"]
 
-    def get(self, frame_to, frame_from=None):
+    def get(
+        self, frame_to: Hashable, frame_from: Optional[Hashable] = None
+    ) -> Tuple[NDArray[np.float64], Optional[Hashable]]:
         """
         Get the transform from one frame to another.
 
@@ -102,6 +110,8 @@ class SceneGraph(object):
         ----------
         transform : (4, 4) float
           Homogeneous transformation matrix
+        geometry
+          The name of the geometry if it exists
 
         Raises
         -----------
@@ -119,8 +129,7 @@ class SceneGraph(object):
             return self._cache[key]
 
         # get the geometry at the final node if any
-        geometry = self.transforms.node_data[
-            frame_to].get('geometry')
+        geometry = self.transforms.node_data[frame_to].get("geometry")
 
         # get a local reference to edge data
         data = self.transforms.edge_data
@@ -130,13 +139,12 @@ class SceneGraph(object):
             matrix = _identity
         elif key in data:
             # if the path is just an edge return early
-            matrix = data[key]['matrix']
+            matrix = data[key]["matrix"]
         else:
             # we have a 3+ node path
             # get the path from the forest always going from
             # parent -> child -> child
-            path = self.transforms.shortest_path(
-                frame_from, frame_to)
+            path = self.transforms.shortest_path(frame_from, frame_to)
             # the path should always start with `frame_from`
             assert path[0] == frame_from
             # and end with the `frame_to` node
@@ -147,21 +155,19 @@ class SceneGraph(object):
             for u, v in zip(path[:-1], path[1:]):
                 forward = data.get((u, v))
                 if forward is not None:
-                    if 'matrix' in forward:
+                    if "matrix" in forward:
                         # append the matrix from u to v
-                        matrices.append(forward['matrix'])
+                        matrices.append(forward["matrix"])
                     continue
                 # since forwards didn't exist backward must
                 # exist otherwise this is a disconnected path
                 # and we should raise an error anyway
                 backward = data[(v, u)]
-                if 'matrix' in backward:
+                if "matrix" in backward:
                     # append the inverted backwards matrix
-                    matrices.append(
-                        np.linalg.inv(backward['matrix']))
+                    matrices.append(np.linalg.inv(backward["matrix"]))
             # filter out any identity matrices
-            matrices = [m for m in matrices if
-                        np.abs((m - _identity)).max() > 1e-8]
+            matrices = [m for m in matrices if np.abs(m - _identity).max() > 1e-8]
             if len(matrices) == 0:
                 matrix = _identity
             elif len(matrices) == 1:
@@ -169,8 +175,14 @@ class SceneGraph(object):
             else:
                 # multiply matrices into single transform
                 matrix = util.multi_dot(matrices)
+
+        # if instructed to repair rigid transforms do it here
+        if self.repair_rigid is not None:
+            matrix = fix_rigid(matrix, max_deviance=self.repair_rigid)
+
         # matrix being edited in-place leads to subtle bugs
-        matrix.flags['WRITEABLE'] = False
+        matrix.flags["WRITEABLE"] = False
+
         # store the result
         self._cache[key] = (matrix, geometry)
 
@@ -210,11 +222,9 @@ class SceneGraph(object):
             if node == base_frame:
                 continue
             # get the matrix and geometry name
-            matrix, geometry = self.get(
-                frame_to=node, frame_from=base_frame)
+            matrix, geometry = self.get(frame_to=node, frame_from=base_frame)
             # store matrix as list rather than numpy array
-            flat[node] = {'transform': matrix.tolist(),
-                          'geometry': geometry}
+            flat[node] = {"transform": matrix.tolist(), "geometry": geometry}
 
         return flat
 
@@ -239,8 +249,7 @@ class SceneGraph(object):
         if mesh_index is None:
             # geometry is an OrderedDict
             # map mesh name to index: {geometry key : index}
-            mesh_index = {name: i for i, name
-                          in enumerate(scene.geometry.keys())}
+            mesh_index = {name: i for i, name in enumerate(scene.geometry.keys())}
 
         # get graph information into local scope before loop
         graph = self.transforms
@@ -251,7 +260,7 @@ class SceneGraph(object):
 
         # list of dict, in gltf format
         # start with base frame as first node index
-        result = [{'name': base_frame}]
+        result = [{"name": base_frame}]
         # {node name : node index in gltf}
         lookup = {base_frame: 0}
 
@@ -262,51 +271,65 @@ class SceneGraph(object):
             # assign the index to the node-name lookup
             lookup[node] = len(result)
             # populate a result at the correct index
-            result.append({'name': node})
+            result.append({"name": node})
 
         # get generated properties outside of loop
         # does the scene have a defined camera to export
         has_camera = scene.has_camera
         children = graph.children
 
+        extensions_used = set()
+
         # then iterate through to collect data
         for info in result:
             # name of the scene node
-            node = info['name']
+            node = info["name"]
 
             # get the original node names for children
             childs = children.get(node, [])
             if len(childs) > 0:
-                info['children'] = [lookup[k] for k in childs]
+                info["children"] = [lookup[k] for k in childs]
 
             # if we have a mesh store by index
-            if 'geometry' in node_data[node]:
-                mesh_key = node_data[node]['geometry']
+            if "geometry" in node_data[node]:
+                mesh_key = node_data[node]["geometry"]
                 if mesh_key in mesh_index:
-                    info['mesh'] = mesh_index[mesh_key]
+                    info["mesh"] = mesh_index[mesh_key]
             # check to see if we have camera node
             if has_camera and node == scene.camera.name:
-                info['camera'] = 0
+                info["camera"] = 0
 
             if node != base_frame:
                 parent = graph.parents[node]
+                node_edge = edge_data[(parent, node)]
 
                 # get the matrix from this edge
-                matrix = edge_data[(parent, node)]['matrix']
+                matrix = node_edge["matrix"]
                 # only include if it's not an identify matrix
                 if not util.allclose(matrix, _identity):
-                    info['matrix'] = matrix.T.reshape(-1).tolist()
+                    info["matrix"] = matrix.T.reshape(-1).tolist()
 
                 # if an extra was stored on this edge
-                extras = edge_data[(parent, node)].get('extras')
+                extras = node_edge.get("metadata")
                 if extras:
+                    extras = extras.copy()
+
+                    # if extensionss were stored on this edge
+                    extensions = extras.pop("gltf_extensions", None)
+                    if isinstance(extensions, dict):
+                        info["extensions"] = extensions
+                        extensions_used = extensions_used.union(set(extensions.keys()))
+
                     # convert any numpy arrays to lists
                     extras.update(
-                        {k: v.tolist() for k, v in extras.items()
-                         if hasattr(v, 'tolist')})
-                    info['extras'] = extras
+                        {k: v.tolist() for k, v in extras.items() if hasattr(v, "tolist")}
+                    )
+                    info["extras"] = extras
 
-        return {'nodes': result}
+        gltf = {"nodes": result}
+        if len(extensions_used) > 0:
+            gltf["extensionsUsed"] = list(extensions_used)
+        return gltf
 
     def to_edgelist(self):
         """
@@ -333,12 +356,12 @@ class SceneGraph(object):
             # make sure we're not stomping on original
             attr_new = attr.copy()
             # apply node geometry to edge attributes
-            if 'geometry' in b_attr:
-                attr_new['geometry'] = b_attr['geometry']
+            if "geometry" in b_attr:
+                attr_new["geometry"] = b_attr["geometry"]
             # convert any numpy arrays to regular lists
             attr_new.update(
-                {k: v.tolist() for k, v in attr_new.items()
-                 if hasattr(v, 'tolist')})
+                {k: v.tolist() for k, v in attr_new.items() if hasattr(v, "tolist")}
+            )
             export.append([a, b, attr_new])
         return export
 
@@ -366,8 +389,7 @@ class SceneGraph(object):
                 self.update(edge[1], edge[0])
             # edge is broken
             elif strict:
-                raise ValueError(
-                    'edge incorrect shape: %s', str(edge))
+                raise ValueError("edge incorrect shape: %s", str(edge))
 
     def to_networkx(self):
         """
@@ -379,9 +401,8 @@ class SceneGraph(object):
           Directed graph.
         """
         import networkx
-        return networkx.from_edgelist(
-            self.to_edgelist(),
-            create_using=networkx.DiGraph)
+
+        return networkx.from_edgelist(self.to_edgelist(), create_using=networkx.DiGraph)
 
     def show(self, **kwargs):
         """
@@ -393,16 +414,14 @@ class SceneGraph(object):
         kwargs : dict
           Passed to `networkx.draw_networkx`
         """
-        import networkx
         import matplotlib.pyplot as plt
+        import networkx
+
         # default kwargs will only be set if not
         # passed explicitly to the show command
-        defaults = {'with_labels': True}
-        kwargs.update(**{k: v for k, v in defaults.items()
-                         if k not in kwargs})
-        networkx.draw_networkx(
-            G=self.to_networkx(),
-            **kwargs)
+        defaults = {"with_labels": True}
+        kwargs.update(**{k: v for k, v in defaults.items() if k not in kwargs})
+        networkx.draw_networkx(G=self.to_networkx(), **kwargs)
 
         plt.show()
 
@@ -440,9 +459,7 @@ class SceneGraph(object):
         nodes_geometry : (m,) array
           Node names which have geometry associated
         """
-        return [n for n, attr in
-                self.transforms.node_data.items()
-                if 'geometry' in attr]
+        return [n for n, attr in self.transforms.node_data.items() if "geometry" in attr]
 
     @caching.cache_decorator
     def geometry_nodes(self):
@@ -457,11 +474,11 @@ class SceneGraph(object):
         """
         res = collections.defaultdict(list)
         for node, attr in self.transforms.node_data.items():
-            if 'geometry' in attr:
-                res[attr['geometry']].append(node)
+            if "geometry" in attr:
+                res[attr["geometry"]].append(node)
         return res
 
-    def remove_geometries(self, geometries):
+    def remove_geometries(self, geometries: Union[str, set, Sequence]):
         """
         Remove the reference for specified geometries
         from nodes without deleting the node.
@@ -472,31 +489,34 @@ class SceneGraph(object):
           Name of scene.geometry to dereference.
         """
         # make sure we have a set of geometries to remove
-        if util.is_string(geometries):
+        if isinstance(geometries, str):
             geometries = [geometries]
         geometries = set(geometries)
 
         # remove the geometry reference from the node without deleting nodes
         # this lets us keep our cached paths, and will not screw up children
-        for node, attrib in self.transforms.node_data.items():
-            if 'geometry' in attrib and attrib['geometry'] in geometries:
-                attrib.pop('geometry')
+        for attrib in self.transforms.node_data.values():
+            if "geometry" in attrib and attrib["geometry"] in geometries:
+                attrib.pop("geometry")
 
         # it would be safer to just run _cache.clear
         # but the only property using the geometry should be
         # nodes_geometry: if this becomes not true change this to clear!
-        self._cache.cache.pop('nodes_geometry', None)
+        self._cache.cache.pop("nodes_geometry", None)
+        self.transforms._hash = None
 
-    def __contains__(self, key):
+    def __contains__(self, key: Hashable) -> bool:
         return key in self.transforms.node_data
 
-    def __getitem__(self, key):
+    def __getitem__(
+        self, key: Hashable
+    ) -> Tuple[NDArray[np.float64], Optional[Hashable]]:
         return self.get(key)
 
-    def __setitem__(self, key, value):
-        value = np.asanyarray(value)
+    def __setitem__(self, key: Hashable, value: ArrayLike):
+        value = np.asanyarray(value, dtype=np.float64)
         if value.shape != (4, 4):
-            raise ValueError('Matrix must be specified!')
+            raise ValueError("Matrix must be specified!")
         return self.update(key, matrix=value)
 
     def clear(self):
@@ -504,7 +524,7 @@ class SceneGraph(object):
         self._cache.clear()
 
 
-class EnforcedForest(object):
+class EnforcedForest:
     """
     A simple forest graph data structure: every node
     is allowed to have exactly one parent. This makes
@@ -547,7 +567,7 @@ class EnforcedForest(object):
         --------
         changed : bool
           Return if this operation changed anything.
-       """
+        """
         self._hash = None
 
         # topology has changed so clear cache
@@ -556,11 +576,9 @@ class EnforcedForest(object):
         else:
             # check to see if matrix and geometry are identical
             edge = self.edge_data[(u, v)]
-            if (util.allclose(kwargs.get('matrix', _identity),
-                              edge.get('matrix', _identity),
-                              1e-8)
-                and (edge.get('geometry') ==
-                     kwargs.get('geometry'))):
+            if util.allclose(
+                kwargs.get("matrix", _identity), edge.get("matrix", _identity), 1e-8
+            ) and (edge.get("geometry") == kwargs.get("geometry")):
                 return False
 
         # store a parent reference for traversal
@@ -569,9 +587,8 @@ class EnforcedForest(object):
         self.edge_data[(u, v)] = kwargs
         # set empty node data
         self.node_data[u].update({})
-        if 'geometry' in kwargs:
-            self.node_data[v].update(
-                {'geometry': kwargs['geometry']})
+        if "geometry" in kwargs:
+            self.node_data[v].update({"geometry": kwargs["geometry"]})
         else:
             self.node_data[v].update({})
 
@@ -600,10 +617,7 @@ class EnforcedForest(object):
         self._hash = None
 
         # delete all children's references and parent reference
-        children = [
-            child for (
-                child,
-                parent) in self.parents.items() if parent == u]
+        children = [child for (child, parent) in self.parents.items() if parent == u]
         for c in children:
             del self.parents[c]
         if u in self.parents:
@@ -673,10 +687,9 @@ class EnforcedForest(object):
                 # we have a either a common node between both
                 # traversal directions or we have consumed the whole
                 # tree in both directions so try to find the common node
-                common = set(backward).intersection(
-                    forward).difference({None})
+                common = set(backward).intersection(forward).difference({None})
                 if len(common) == 0:
-                    raise ValueError('No path from {}->{}!'.format(u, v))
+                    raise ValueError(f"No path from {u}->{v}!")
                 elif len(common) > 1:
                     # get the first occurring common element in "forward"
                     link = next(f for f in forward if f in common)
@@ -686,8 +699,8 @@ class EnforcedForest(object):
                     link = next(iter(common))
 
                 # combine the forward and backwards traversals
-                a = forward[:forward.index(link) + 1]
-                b = backward[:backward.index(link)]
+                a = forward[: forward.index(link) + 1]
+                b = backward[: backward.index(link)]
                 path = a + b[::-1]
 
                 # verify we didn't screw up the order
@@ -698,7 +711,7 @@ class EnforcedForest(object):
 
                 return path
 
-        raise ValueError('Iteration limit exceeded!')
+        raise ValueError("Iteration limit exceeded!")
 
     @property
     def nodes(self):
@@ -722,14 +735,16 @@ class EnforcedForest(object):
         children : dict
           Keyed {node : [child, child, ...]}
         """
+        if "children" in self._cache:
+            return self._cache["children"]
         child = collections.defaultdict(list)
         # append children to parent references
         # skip self-references to avoid a node loop
-        [child[v].append(u) for u, v in
-         self.parents.items() if u != v]
+        [child[v].append(u) for u, v in self.parents.items() if u != v]
 
-        # return as a vanilla dict
-        return dict(child)
+        # cache and return as a vanilla dict
+        self._cache["children"] = dict(child)
+        return self._cache["children"]
 
     def successors(self, node):
         """
@@ -750,7 +765,7 @@ class EnforcedForest(object):
         children = self.children
         # if node doesn't exist return early
         if node not in children:
-            return set([node])
+            return {node}
 
         # children we need to collect
         queue = [node]
@@ -771,38 +786,44 @@ class EnforcedForest(object):
 
     def __hash__(self):
         """
-        Actually hash all of the data.
-
-        Previously we were relying on "dirty" flags but
-        that made the bookkeeping unreasonably critical.
+        Actually hash all of the data, but use a "dirty" mechanism
+        in functions that modify the data, which MUST
+        # all invalidate the hash by setting `self._hash = None`
 
         This was optimized a bit, and is evaluating on an
         older laptop on a scene with 77 nodes and 76 edges
         10,000 times in 0.7s which seems fast enough.
         """
-        hashed = getattr(self, '_hash', None)
+        # see if there is an available hash value
+        # if you are seeing cache bugs this is the thing
+        # to try eliminating because it is very likely that
+        # someone somewhere is modifying the data without
+        # setting `self._hash = None`
+        hashed = getattr(self, "_hash", None)
         if hashed is not None:
             return hashed
 
         hashed = hash_fast(
-            (''.join(str(hash(k)) + v.get('geometry', '')
-                     for k, v in self.edge_data.items()) +
-             ''.join(str(k) + v.get('geometry', '')
-                     for k, v in self.node_data.items())).encode('utf-8') +
-            b''.join(v['matrix'].tobytes()
-                     for v in self.edge_data.values()
-                     if 'matrix' in v))
+            (
+                "".join(
+                    str(hash(k)) + v.get("geometry", "")
+                    for k, v in self.edge_data.items()
+                )
+                + "".join(
+                    str(k) + v.get("geometry", "") for k, v in self.node_data.items()
+                )
+            ).encode("utf-8")
+            + b"".join(
+                v["matrix"].tobytes() for v in self.edge_data.values() if "matrix" in v
+            )
+        )
         self._hash = hashed
         return hashed
 
 
 def kwargs_to_matrix(
-        matrix=None,
-        quaternion=None,
-        translation=None,
-        axis=None,
-        angle=None,
-        **kwargs):
+    matrix=None, quaternion=None, translation=None, axis=None, angle=None, **kwargs
+):
     """
     Take multiple keyword arguments and parse them
     into a homogeneous transformation matrix.
@@ -816,9 +837,9 @@ def kwargs_to_matrix(
         # a matrix takes immediate precedence over other options
         return np.array(matrix, dtype=np.float64)
     elif quaternion is not None:
-        matrix = transformations.quaternion_matrix(quaternion)
+        matrix = quaternion_matrix(quaternion)
     elif axis is not None and angle is not None:
-        matrix = transformations.rotation_matrix(angle, axis)
+        matrix = rotation_matrix(angle, axis)
     else:
         matrix = np.eye(4)
 
